<?php

namespace Drupal\ai_agent_migration\Plugin\AiAgent;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ai_agents\Attribute\AiAgent;
use Drupal\ai_agents\Exception\AgentProcessingException;
use Drupal\ai_agents\Plugin\AiAgent\ContentType;
use Drupal\ai_agents\PluginInterfaces\AiAgentInterface;
use Drupal\node\Entity\Node;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the Node Content Type Agent.
 */
#[AiAgent(
  id: 'migration_agent',
  label: new TranslatableMarkup('Migration Agent'),
)]
class Migration extends ContentType {

  use DependencySerializationTrait;
  /**
   * The HTTP client.
   *
   * @var \Drupal\Core\Http\ClientInterface
   */
  protected $httpClient;

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $parent_instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $parent_instance->entityFieldManager = $container->get('entity_field.manager');
    $parent_instance->httpClient = $container->get('http_client');
    return new static($parent_instance);
  }

  /**
   * {@inheritDoc}
   */
  public function getId() {
    return 'migration_agent';
  }

  /**
   * {@inheritDoc}
   */
  public function agentsNames() {
    return [
      'Migration Agent',
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function agentsCapabilities() {
    return [
      'migration_agent' => [
        'name' => 'Migration Agent',
        'description' => "This is capable of migrating a list of URLs and also answer questions about migrations.",
        'inputs' => [
          'free_text' => [
            'name' => 'Prompt',
            'type' => 'string',
            'description' => 'The prompt to create, edit, delete or ask questions about migrations.',
            'default_value' => '',
          ],
        ],
        'outputs' => [
          'answers' => [
            'description' => 'The answers to the questions asked about the migration generated.',
            'type' => 'string',
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function setData($data) {
    $this->data[] = $data;
  }

  /**
   * {@inheritDoc}
   */
  public function getData() {
    return $this->data;
  }

  /**
   * {@inheritDoc}
   */
  public function answerQuestion() {
    $data = $this->agentHelper->runSubAgent('answerQuestion', [
      'The list of migrations' => $this->getVerboseNodeTypesAsString(),
    ]);

    $answer = "";
    if (isset($data[0]['answer'])) {
      foreach ($data as $dataPoint) {
        $answer .= $dataPoint['answer'] . "\n";
      }
      return $answer;
    }

    return $this->t("Sorry, I've got no answers for you.");
  }

  /**
   * {@inheritDoc}
   */
  public function getHelp() {
    $help = $this->t("This agent can migrate a list of URLs.");
    return $help;
  }

  /**
   * {@inheritDoc}
   */
  public function hasAccess() {
    // Check for permissions.
    if (!$this->currentUser->hasPermission('create page content')) {
      return AccessResult::forbidden();
    }
    return parent::hasAccess();
  }

  /**
   * {@inheritDoc}
   */
  public function determineSolvability() {
    parent::determineSolvability();
    $this->taskType = $this->determineTypeOfTask();
    switch ($this->taskType) {
      case 'migrate':
        return AiAgentInterface::JOB_SOLVABLE;

      case 'information':
        return AiAgentInterface::JOB_SHOULD_ANSWER_QUESTION;

      case 'fail':
        return AiAgentInterface::JOB_INFORMS;
    }

    return AiAgentInterface::JOB_NOT_SOLVABLE;
  }

  /**
   * {@inheritDoc}
   */
  public function askQuestion() {
    return $this->questions;
  }

  /**
   * {@inheritDoc}
   */
  public function solve() {
    parent::solve();
    switch ($this->data[0]['action']) {
      case 'migrate':
        $this->executeMigrate();
        break;

      default:
        $this->result[] = 'We could not figure out what you wanted to do.';
    }
    return implode("\n", $this->result);
  }

  /**
   * {@inheritDoc}
   */
  public function approveSolution() {
    $this->data[0]['action'] = 'migrate';
  }

  /**
   * Check so all requirements are there.
   *
   * @return bool
   *   If all requirements are there.
   */
  public function checkRequirements() {
    return TRUE;
  }

  /**
   * Determine if the context is asking a question or wants a audit done.
   *
   * @return string
   *   The context.
   */
  public function determineTypeOfTask() {
    $data = $this->agentHelper->runSubAgent('determineMigrationTask', []);

    if (isset($data[0]['action']) && in_array($data[0]['action'], [
      'migrate',
      'question',
      'information',
      'fail',
    ])) {
      // Inform the user of failure.
      if ($data[0]['action'] === 'fail') {
        $this->information = $data[0]['fail_message'];
      }

      $this->data = $data;
      return $data[0]['action'];
    }
    throw new \Exception('The migration agent can not understand your request.');
  }

  /**
   * Migrate.
   */
  public function executeMigrate() {
    // Collect the context values.
    $urls = $this->data[1]['urls'];
    $node_type = $this->data[2]['node_type'] ?? 'page';
    if (!empty($urls)) {
      foreach ($urls as $url) {
        stripslashes($url);
        // Do some waiting to not overwhelm.
        sleep(1);
        // Get URL.
        $page = $this->getUrl($url);
        if (empty($page)) {
          $this->result[] = "No results found for the search query: " . $url;
        }
        else {
          // Create a new node object.
          $node = Node::create([
            'type' => $node_type,
            // Node field values.
            'title' => $page['title'],
            'field_content' => [
              'value' => $page['body'],
              'format' => 'content_format',
            ],
          ]);
          if ($node->save()) {
            // Check for success.
            if ($node->id()) {
              $this->result[] = $this->t("%node_type node '%title' created: @link.", [
                '%node_type' => ucwords(strtolower($node_type)),
                '%title' => $page['title'],
                '@link' => Url::fromRoute('entity.node.canonical', ['node' => $node->id()])->toString(),
              ]);
            }
            else {
              $this->result[] = "Node creation failed.";
            }
          }
          else {
            throw new AgentProcessingException('There was an error creating the node.');
          }
        }
      }
    }
    else {
      $this->result[] = "No URLs found.";
    }

    return $this->result;
  }

  /**
   * Get the URL.
   *
   * @param string $url
   *   The URL of the article to get.
   *
   * @return array
   *   The article as an associative array with 'title' and 'body'.
   */
  public function getUrl(string $url): array {
    $request = $this->httpClient->request('GET', $url);
    if ($request->getStatusCode() !== 200) {
      throw new AgentProcessingException('There was an error getting the URL.');
    }
    $contents = $request->getBody()->getContents();
    preg_match("'<title[^>]*>(.*?)</title>'s", $contents, $match);
    $title = $match[1];
    // Get the title of the page.
    if (empty($title)) {
      $title = 'No title found: ' . $url;
    }
    // Remove unwanted tags.
    $body = $this->removeTags($contents);
    $body = strip_tags($body, '<p><a><h1><h2><h3><h4><h5><h6><ul><ol><li><br><strong><em><b><i>');
    $markdown = [
      'title' => $title,
      'body' => $body,
    ];
    return $markdown;
  }

  /**
   * Removes <script>/<style> tags and their content from the provided string.
   *
   * This function uses regular expressions to strip out all <script> and
   * <style> tags, including their contents, from the given HTML content. It is
   * useful for sanitizing HTML by removing potentially unsafe or unnecessary
   * scripts and styles.
   *
   * @param string $content
   *   The HTML content to be sanitized.
   *
   * @return string
   *   The sanitized HTML content with <script> and <style> tags removed.
   */
  public function removeTags(string $content): string {
    $content = preg_replace('/<script\b[^>]*>(.*?)<\/script>/is', '', $content);
    $content = preg_replace('/<style\b[^>]*>(.*?)<\/style>/is', '', $content);
    return $content;
  }

}
